@@ -15,6 +15,7 @@ from kabaret.app.ui.gui.widgets.flow.flow_view import (
 from kabaret.app.ui.gui.widgets.editors import editor_factory
 from kabaret.flow_contextual_dict import ContextualView, get_contextual_dict
 from kabaret.subprocess_manager.runners import Explorer
+from kabaret.flow_entities.store import EntityStore
 
 from .users import Users, UserEnvironment, UserBookmarks, UserProfile
 from .film import Sequences
@@ -40,7 +41,7 @@ from .site import (
     MinioFileUploader, MinioFileDownloader,
     RequestRevisions, MultiRequestRevisions,
 )
-from .file import TrackedFile, Revision
+from .file import TrackedFile, Revision, FileManager
 from .kitsu import KitsuAPIWrapper, KitsuBindings
 from .dependency import DependencyTemplates
 
@@ -130,7 +131,8 @@ class SynchronizeFilesResult(flow.Action):
 
 
 class SynchronizeFiles(Synchronize):
-    ICON = ("icons.libreflow", "sync_arrow")
+    
+    ICON = ('icons.libreflow', 'sync')
 
     result = flow.Child(SynchronizeFilesResult)
 
@@ -311,7 +313,9 @@ class Synchronization(flow.Object):
         return jobs.count_waiting.get()
 
     def summary(self):
-        nb_waiting_jobs = self._ensure_waiting_count()
+        nb_waiting_jobs = self.root().project().get_current_site().count_jobs(
+            status='WAITING'
+        )
         
         if nb_waiting_jobs > 0:
             return (
@@ -360,14 +364,39 @@ class KitsuConfig(flow.Object):
     gazu_api = flow.Child(KitsuAPIWrapper).ui(
         hidden=True
     )
+    configured = flow.Computed().ui(
+        editor='bool',
+        hidden=True
+    )
 
     def child_value_changed(self, child_value):
         self.project_id.touch()
 
     def compute_child_value(self, child_value):
         if child_value is self.project_id:
-            data = self.gazu_api.get_project_data()
-            child_value.set(data["id"])
+            self.project_id.set(
+                self.gazu_api.get_project_id()
+            )
+        elif child_value is self.configured:
+            self.configured.set(
+                self.project_id.get() is not None
+            )
+    
+    def is_uploadable(self, file_name):
+        patterns = self.uploadable_files.get().split(',')
+
+        for pattern in patterns:
+            pattern = pattern.replace(' ', '')
+            if not fnmatch.fnmatch(file_name, pattern):
+                return False
+    
+    def summary(self):
+        if not self.configured.get():
+            return (
+                '<font color=#D5000D>'
+                '<b>Server URL</b> and <b>project name</b> '
+                'unset: Kitsu features won\'t be available'
+            )
 
 
 class DefaultFiles(flow.Object):
@@ -383,10 +412,14 @@ class DefaultFilesMap(flow.Map):
 
 
 class ProjectSettings(flow.Object):
+
+    ICON = ('icons.gui', 'settings')
+
     project_nice_name = flow.Param()
     project_thumbnail = flow.Param().ui(editor='textarea')
     support = flow.Param("")
 
+    settings = flow.Child(ContextualView)
     default_files = flow.Child(DefaultFilesMap)
     non_editable_files = flow.Param("")
     auto_upload = flow.Param('').ui(
@@ -485,13 +518,21 @@ class MultisiteConfig(flow.Object):
         return data
 
     def summary(self):
+        current_site = self.working_sites[self.current_site_name.get()]
         exchange_site = self.exchange_sites[self.exchange_site_name.get()]
+
+        if not current_site.configured.get():
+            return (
+                '<font color=#D5000D>'
+                f'Current site <b>{current_site.name()}</b> '
+                'needs to be configured</font>'
+            )
         
         if not exchange_site.configured.get():
             return (
-                "<font color=#D66700>"
-                "Exchange site <b>%s</b> not completely configured."
-                "</font>" % exchange_site.name()
+                '<font color=#D5000D>'
+                f'Exchange site <b>{exchange_site.name()}</b> '
+                'needs to be configured.</font>'
             )
 
 
@@ -502,19 +543,22 @@ class Admin(flow.Object):
     _project = flow.Parent()
 
     project_settings = flow.Child(ProjectSettings).injectable()
-    settings = flow.Child(ContextualView)
-    users = flow.Child(Users)
+    users = flow.Child(Users).ui(show_filter=True)
+    multisites = flow.Child(MultisiteConfig).ui(label='Sites')
     kitsu = flow.Child(KitsuConfig)
 
     dependency_templates = flow.Child(DependencyTemplates)
     
-    user_environment = flow.Child(UserEnvironment).ui(expanded=False)
+    user_environment = flow.Child(UserEnvironment).ui(
+        expanded=False,
+        label='Session environment')
     default_applications = flow.Child(DefaultRunners).ui(expanded=False).injectable()
 
-    multisites = flow.Child(MultisiteConfig).ui(label="Multi-site configuration")
-    process_jobs = flow.Child(Synchronize)
     login_page = flow.Child(LoginPage).ui(hidden=True)
 
+    file_manager = flow.Child(FileManager)
+    store = flow.Child(EntityStore)
+
 
 class Project(flow.Object):
 
@@ -540,7 +584,10 @@ class Project(flow.Object):
     def get_temp_folder(self):
         return self.admin.multisites.temp_dir.get()
 
-    def set_user(self, username):
+    def set_user_name(self, username):
+        '''
+        Sets the name of the current user.
+        '''
         project_settings_folder = self.project_settings_folder()
 
         if not os.path.exists(project_settings_folder):
@@ -554,10 +601,22 @@ class Project(flow.Object):
         
         self.user.current_user_id.touch()
 
-    def get_user(self):
+    def get_user_name(self):
+        '''
+        Returns the name of the current user.
+        '''
         return self.user.current_user_id.get()
-    
-    def get_user_from_name(self, name):
+
+    def get_user(self, name=None):
+        '''
+        If provided, returns the user of this project with the
+        given `name`. Return the project's current user otherwise.
+
+        In any case, None is returned if no user is found.
+        '''
+        if name is None:
+            name = self.get_user_name()
+        
         if not self.admin.users.has_mapped_name(name):
             return None
         
@@ -653,18 +712,18 @@ class Project(flow.Object):
 
         # Set current user
         # user_name = self.admin.users.get_user_name(kitsu_login)
-        user = self.admin.users.get_user(kitsu_id)
+        user = self.admin.users.get_user_by_uid(kitsu_id)
 
         if user is None:
             # Create user profile if not registered
             user = self.admin.users.add_user(kitsu_id, kitsu_login)
 
-            if not user:
+            if user is None:
                 return 2
 
             # user_name = user.name()
         
-        self.set_user(user.name())
+        self.set_user_name(user.name())
 
         # Create user folder at first login
         user_settings_folder = self.user_settings_folder()
@@ -692,6 +751,9 @@ class Project(flow.Object):
     def kitsu_name(self):
         return self.admin.kitsu.project_name.get()
     
+    def kitsu_config(self):
+        return self.admin.kitsu
+    
     def kitsu_api(self):
         return self.admin.kitsu.gazu_api
 
@@ -701,7 +763,7 @@ class Project(flow.Object):
 
     def get_contextual_view(self, context_name):
         if context_name == "settings":
-            return self.admin.settings
+            return self.admin.project_settings.settings
 
     def get_default_contextual_edits(self, context_name):
         if context_name == "settings":
@@ -714,13 +776,6 @@ class Project(flow.Object):
                 project_thumbnail="{ROOT_DIR}TECH/{PROJECT}_thumbnail.png",
             )
 
-    def get_user_bookmarks(self):
-        user = self.get_user()
-        if user in self.admin.users.mapped_names():
-            return self.admin.users[user].bookmarks
-        else:
-            return None
-
     def project_settings_folder(self):
         return os.path.join(
             pathlib.Path.home(),
@@ -734,7 +789,7 @@ class Project(flow.Object):
         '''
         return os.path.join(
             self.project_settings_folder(),
-            self.get_user()
+            self.get_user_name()
         )
 
     def get_project_thumbnail2(self):
@@ -792,6 +847,15 @@ class Project(flow.Object):
     def get_factory(self):
         self.ensure_runners_loaded()
         return self._RUNNERS_FACTORY
+
+    def get_entity_store(self):
+        return self.admin.store
+    
+    def get_file_manager(self):
+        return self.admin.file_manager
+    
+    def get_sync_manager(self):
+        return self.synchronization.synchronize_files
     
     def touch(self):
         super(Project, self).touch()
