@@ -194,7 +194,13 @@ class Department(flow.Object):
     files = (
         flow.Child(FileSystemMap)
         .injectable()
-        .ui(expanded=True, action_submenus=True, items_action_submenus=True)
+        .ui(
+            expanded=True,
+            # auto_fit=False,
+            # columns_width=(10, 40, 50),
+            action_submenus=True,
+            items_action_submenus=True
+        )
     )
     auto_current = flow.BoolParam(True).ui(hidden=True)
 
