@@ -11,10 +11,13 @@ import fnmatch
 import re
 import traceback
 import tempfile
+import uuid
+import pathlib
 
 from kabaret import flow
 from kabaret.flow_contextual_dict import ContextualView, get_contextual_dict
 from kabaret.subprocess_manager.flow import RunAction
+from kabaret.flow_entities.entities import EntityCollection, Entity, Property, PropertyValue
 
 from ..utils.os import zip_folder, unzip_archive
 
@@ -35,15 +38,23 @@ class StaticSiteSyncTransferStateChoices(flow.values.ChoiceValue):
     CHOICES = ['NC', 'UP_REQUESTED', 'DL_REQUESTED', 'UP_IN_PROGRESS', 'DL_IN_PROGRESS']
 
 
-class LoadType(flow.values.ChoiceValue):
+class LoadType(PropertyValue):
 
+    DEFAULT_EDITOR = 'choice'
     CHOICES = ['Upload', 'Download']
 
+    def choices(self):
+        return self.CHOICES
 
-class JobStatus(flow.values.ChoiceValue):
 
+class JobStatus(PropertyValue):
+
+    DEFAULT_EDITOR = 'choice'
     CHOICES = ['WFA', 'WAITING', 'PROCESSED', 'ERROR', 'PAUSE', 'DONE']
 
+    def choices(self):
+        return self.CHOICES
+
 
 class ResetJob(flow.Action):
 
@@ -112,18 +123,20 @@ class ShowRevisionInHistory(flow.Action):
         return self.get_result(goto=history_oid)
 
 
-class Job(flow.Object):
+class Job(Entity):
 
-    type = flow.Param('Download', LoadType)
-    status = flow.Param('WAITING', JobStatus)
-    priority = flow.IntParam(50)
-    emitter_oid = flow.Param()
-    date = flow.Param("").ui(editor='datetime')
-    date_end = flow.Param("").ui(editor='datetime')
-    is_archived = flow.BoolParam()
-    requested_by_user = flow.Param()
-    requested_by_studio = flow.Param()
-    log = flow.Param().ui(editor='textarea')
+    type                = Property(property_value_type=LoadType)
+    status              = Property(property_value_type=JobStatus)
+    priority            = Property().ui(editor='int')
+    emitter_oid         = Property()
+    local_path          = Property()
+    server_path         = Property()
+    date                = Property().ui(editor='datetime')
+    date_end            = Property().ui(editor='datetime')
+    is_archived         = Property().ui(editor='bool')
+    requested_by_user   = Property()
+    requested_by_studio = Property()
+    log                 = Property().ui(editor='textarea')
 
     reset = flow.Child(ResetJob)
     show = flow.Child(ShowRevisionInHistory).ui(label='Show in history')
@@ -143,9 +156,69 @@ class Job(flow.Object):
             )
 
         return job_repr
+    
+    def on_submitted(self):
+        '''
+        Computes the local and server paths of the job's
+        requested object (identified by `emitter_oid`).
+
+        If the oid does not refer to an existing revision
+        object, the methods puts the job in error state.
+        '''
+        try:
+            revision = self.root().get_object(
+                self.emitter_oid.get()
+            )
+        except (ValueError, flow.exceptions.MissingRelationError):
+            self.status.set('ERROR')
+            self.log.set('Object does not exist')
+            return
+        
+        try:
+            ready_for_sync = revision.ready_for_sync.get()
+            local_path = revision.get_path()
+            server_path = revision.get_path(relative=True)
+        except AttributeError:
+            self.status.set('ERROR')
+            self.log.set('Object is not a revision')
+            return
+        else:
+            if not ready_for_sync:
+                self.status.set('ERROR')
+                self.log.set('Revision is not ready to be synced')
+                return
+        
+        if os.path.splitext(server_path)[1] == '':
+            server_path += '/' + os.path.basename(server_path) + '.zip'
+        
+        self.local_path.set(local_path)
+        self.server_path.set(server_path.replace('\\', '/'))
+    
+    def on_processed(self):
+        '''
+        Method which may be called after this job has been
+        processed.
+
+        Default is to update the related revision's statutes,
+        and update its UI and that of its parent map. The
+        method assumes param `emitter_oid` references a valid
+        revision object.
+        '''
+        if self.status.get() != 'PROCESSED':
+            return
+        
+        revision = self.root().get_object(
+            self.emitter_oid.get()
+        )
+        revision.set_sync_status(
+            'Available',
+            exchange=(self.type.get() == 'Upload')
+        )
+        revision.touch()
+        revision._revisions.touch()
 
 
-class JobQueue(flow.Map):
+class JobQueue(EntityCollection):
 
     count_waiting = flow.IntParam(0)
     reset_jobs = flow.Child(ResetJobs).ui(label='Reset erroneous jobs')
@@ -164,21 +237,24 @@ class JobQueue(flow.Map):
     def jobs(self, type=None, status=None):
         jobs = self.mapped_items()
 
-        if type:
-            jobs = list(filter(lambda job: job.type.get() == type, jobs))
-        if status:
-            jobs = list(filter(lambda job: job.status.get() == status, jobs))
+        if type is not None:
+            jobs = [j for j in jobs if j.type.get() == type]
+        if status is not None:
+            jobs = [j for j in jobs if j.status.get() == status]
         
         return jobs
     
-    def insert(self, name, priority=0):
-        """
-        Inserts a new job with given name and priority.
-        """
-        job = self.add(name)
-        self._mapped_names.set_score(name, priority)
+    def count(self, type=None, status=None):
+        job_filter = {}
 
-        return job
+        if type is not None:
+            job_filter['type'] = type
+        if status is not None:
+            job_filter['status'] = status
+        
+        c = self.get_entity_store().get_collection(self.collection_name())
+
+        return c.count_documents(job_filter)
     
     def update_waiting_count(self, count=None):
         if count is None:
@@ -208,8 +284,10 @@ class JobQueue(flow.Map):
             job_type,
             time.time()
         )
-        job = self.insert(name, priority)
+        self.ensure_exist([name])
+        self._document_cache = None
 
+        job = self.get_mapped(name)
         job.type.set(job_type)
         job.status.set(init_status)
         job.priority.set(priority)
@@ -221,6 +299,8 @@ class JobQueue(flow.Map):
         job.is_archived.set(False)
         job.log.set('?')
 
+        job.on_submitted()
+
         self.incr_waiting_count()
 
         self.root().session().log_info("Submitted job %s" % job)
@@ -235,18 +315,22 @@ class JobQueue(flow.Map):
             "Status",
             "Revision",
             "Emitted on",
-            "Requested by (site/user)",
+            "By",
+            "Site",
         ]
     
     def _fill_row_cells(self, row, item):
-        row["Type"] = item.type.get()
-        row["Status"] = item.status.get()
-        row["Revision"] = item.emitter_oid.get()
-        row["Emitted on"] = datetime.fromtimestamp(item.date.get()).ctime()
-        row["Requested by (site/user)"] = "%s/%s" % (
-            item.requested_by_studio.get(),
-            item.requested_by_user.get()
-        )
+        self.mapped_names()
+        item_data = self._document_cache[item.name()]
+
+        row['Type']       = item_data['type']
+        row['Status']     = item_data['status']
+        row['Revision']   = item_data['emitter_oid']
+        row['By']         = item_data['requested_by_user']
+        row['Site']       = item_data['requested_by_studio']
+        row['Emitted on'] = datetime.fromtimestamp(
+            float(item_data['date'])
+        ).ctime()
 
 
 class ProcessJobs(flow.Action):
@@ -258,7 +342,7 @@ class ProcessJobs(flow.Action):
     def needs_dialog(self):
         return False
 
-    def _process(self, job):
+    def process(self, job):
         raise NotImplementedError(
             "Must be implemented to process the given job"
         )
@@ -282,7 +366,7 @@ class ProcessJobs(flow.Action):
             self.root().session().log_info("Processing job %s" % job)
             
             self._job_count[job.status.get()] -= 1
-            self._process(job)
+            self.process(job)
             self._job_count[job.status.get()] += 1
         
         current_site = self.root().project().get_current_site()
@@ -367,137 +451,131 @@ class MinioDownloadFile(flow.Object):
         )
 
 
-class Synchronize(ProcessJobs):
+class SyncManager(flow.Object):
 
-    uploader = flow.Child(MinioUploadFile).ui(hidden=True)
-    downloader = flow.Child(MinioDownloadFile).ui(hidden=True)
+    _exchange = flow.Parent()
 
-    def _get_jobs(self):
-        # Get current site's waiting jobs only
-        current_site = self.root().project().get_current_site()
-        return current_site.get_jobs(status="WAITING")
+    def __init__(self, parent, name):
+        super(SyncManager, self).__init__(parent, name)
+        self._client = None
 
-    def _get_sync_status(self, revision_oid, site_name=None, site_type="Studio"):
-        if not site_name and site_type == "Exchange":
-            exchange_site = self.root().project().get_exchange_site()
-            if exchange_site:
-                site_name = exchange_site.name()
-            else:
-                return "NotAvailable"
-
-        sync_status = self.root().session().cmds.Flow.call(
-            revision_oid, "get_sync_status",
-            args={}, kwargs=dict(site_name=site_name)
+    def upload(self, server_path, local_path):
+        zipped_folder = self._is_zipped_folder(
+            server_path, local_path
         )
-        return sync_status
+        if zipped_folder:
+            # Folder to be zipped before upload
+            src_path = self._get_temp_zip(server_path)
+            zip_folder(local_path, src_path)
+        else:
+            # File/folder uploaded as is
+            src_path = local_path
 
-    def _set_sync_status(self, revision_oid, status, site_name=None, site_type="Studio"):
-        if not site_name and site_type == "Exchange":
-            exchange_site = self.root().project().get_exchange_site()
-            if exchange_site:
-                site_name = exchange_site.name()
-            else:
-                self.root().session().log_error(
-                    "No registered exchange site on this project !"
-                )
-                return
+        self.root().session().log_info(
+            f'Upload file {local_path} -> {server_path}'
+        )
+        self._ensure_client().fput_object(
+            self._exchange.bucket_name.get(),
+            server_path,
+            src_path
+        )
 
-        self.root().session().cmds.Flow.call(
-            revision_oid, "set_sync_status",
-            args={status}, kwargs=dict(site_name=site_name)
+        if zipped_folder:
+            os.remove(src_path)
+
+    def download(self, server_path, local_path):
+        zipped_folder = self._is_zipped_folder(
+            server_path, local_path
+        )
+        if zipped_folder:
+            dst_path = self._get_temp_zip(server_path)
+        else:
+            dst_path = local_path
+        
+        self.root().session().log_info(
+            f'Download file {server_path} -> {local_path}'
+        )
+        self._ensure_client().fget_object(
+            self._exchange.bucket_name.get(),
+            server_path,
+            dst_path
+        )
+
+        if zipped_folder:
+            unzip_archive(dst_path, local_path)
+            os.remove(dst_path)
+    
+    def check_connection(self):
+        try:
+            self._ensure_client().list_buckets()
+        except Exception as err:
+            return str(err)
+        else:
+            return None
+
+    def _ensure_client(self):
+        if self._client is None:
+            self._client = Minio(
+                self._exchange.server_url.get(),
+                access_key=self._exchange.server_login.get(),
+                secret_key=self._exchange.server_password.get(),
+                secure=self._exchange.enable_tls.get()
+            )
+        
+        return self._client
+    
+    def _get_temp_zip(self, server_path):
+        return os.path.join(
+            self.root().project().get_temp_folder(),
+            (
+                pathlib.PurePath(server_path).stem
+                + f'-{str(uuid.uuid4())}.zip'
+            )
         )
     
-    def _touch(self, oid):
-        oid = self.root().session().cmds.Flow.resolve_path(oid)
-        self.root().session().cmds.Flow.call(
-            oid, "touch", args={}, kwargs={}
+    def _is_zipped_folder(self, server_path, local_path):
+        return (
+            os.path.splitext(local_path)[1] == ''
+            and os.path.splitext(server_path)[1] == '.zip'
         )
 
-    def _process(self, job):
-        if job.status.get() == "PROCESSED":
+
+class Synchronize(ProcessJobs):
+
+    def _get_jobs(self):
+        return self.root().project().get_current_site().get_jobs(
+            status='WAITING'
+        )
+
+    def process(self, job):
+        if job.status.get() != 'WAITING':
             self.root().session().log_warning(
-                "Job %s already processed !" % job.name()
+                'Job %s in %s status: skipping' % (
+                    job.name(),
+                    job.status.get()
+                )
             )
             return
-
-        local_path = self.root().session().cmds.Flow.call(
-            job.emitter_oid.get(),
-            "get_path",
-            args={},
-            kwargs={}
-        )
-        server_path = self.root().session().cmds.Flow.call(
-            job.emitter_oid.get(),
-            "get_relative_path",
-            args={},
-            kwargs={}
-        )
-        local_path = local_path.replace("\\", "/")
-        server_path = server_path.replace("\\", "/")
-
-        try:
-            if job.type.get() == "Upload":
-                # Check if local_path refers to a folder (.zip format)
-                local_folder = os.path.dirname(local_path)
-                
-                if os.path.splitext(local_path)[1] == '.zip':
-                    local_folder = os.path.dirname(local_path)
-                    
-                    if not os.path.exists(local_folder):
-                        # Return if folder does not exist
-                        self.root().session().log_error((
-                            "Following folder requested for upload but does not exist !"
-                            "\n>>>>> %s" % local_folder
-                        ))
-                        return
-                    else:
-                        # Zip folder files
-                        zip_folder(local_folder, local_path)
-
-                elif not os.path.exists(local_path):
-                    self.root().session().log_error((
-                        "Following file requested for upload but does not exist !"
-                        "\n>>>>> %s" % local_path
-                    ))
-                    return
-
-                self.uploader.upload(local_path, server_path)
-                
-                if os.path.splitext(local_path)[1] == '.zip':
-                    os.remove(local_path)
-                
-                self._set_sync_status(job.emitter_oid.get(), "Available", site_type="Exchange")
-                self._touch(job.emitter_oid.get() + "/..")
-            elif job.type.get() == "Download":
-                sync_status = self._get_sync_status(job.emitter_oid.get(), site_type="Exchange")
-                
-                if sync_status != "Available":
-                    self.root().session().log_error((
-                        "Following file requested for download "
-                        "but not available on any exchange server !"
-                        "\n>>>>> %s" % local_path
-                    ))
-                    return
-
-                self.downloader.download(server_path, local_path)
-                
-                # If it's a folder revision, extract its archive content
-                if os.path.splitext(local_path)[1] == '.zip':
-                    unzip_archive(local_path, os.path.dirname(local_path))
-                    os.remove(local_path)
-                
-                self._set_sync_status(job.emitter_oid.get(), "Available")
-                self._touch(job.emitter_oid.get() + "/..")
-                self._touch(job.emitter_oid.get() + "/../../../..")
         
-        except Exception as err:
-            job.log.set(traceback.format_exc())
+        try:
+            sync_manager = self.root().project().get_exchange_site().sync_manager
+        
+            if job.type.get() == 'Upload':
+                sync_manager.upload(
+                    job.server_path.get(),
+                    job.local_path.get()
+                )
+            else:
+                sync_manager.download(
+                    job.server_path.get(),
+                    job.local_path.get()
+                )
+        except Exception as e:
             job.status.set('ERROR')
+            job.log.set(traceback.format_exc())
         else:
-            job.status.set("PROCESSED")
-        finally:
-            # Decrement number of waiting jobs
-            self.root().project().get_current_site().queue.job_list.decr_waiting_count()
+            job.status.set('PROCESSED')
+            job.on_processed()
 
 
 class ActiveSiteChoiceValue(flow.values.SessionValue):
@@ -660,7 +738,7 @@ class RequestAs(flow.Action):
             job_type="Download",
             init_status="WAITING",
             emitter_oid=self._revision.oid(),
-            user=self.root().project().get_user(),
+            user=self.root().project().get_user_name(),
             studio=target_site_name,
             priority=self.priority.get(),
         )
@@ -705,7 +783,7 @@ class RequestAs(flow.Action):
             job_type="Upload",
             init_status="WAITING",
             emitter_oid=self._revision.oid(),
-            user=self.root().project().get_user(),
+            user=self.root().project().get_user_name(),
             studio=target_site_name,
             priority=self.priority.get(),
         )
@@ -753,7 +831,8 @@ class Request(RequestAs):
         return (
             context
             and not self._revision.is_working_copy()
-            and not self._revision.get_sync_status() == "Available"
+            and self._revision.get_sync_status() != "Available"
+            and self._revision.get_sync_status(exchange=True) != "Available"
         )
 
 
@@ -922,7 +1001,7 @@ class Site(flow.Object):
     short_name = flow.Param("")
     description = flow.Param("")
     # Define conditions required for a site to be considered as correctly configured
-    configured = flow.Computed()
+    configured = flow.Computed().ui(editor='bool')
     configuration = flow.Child(ConfigureSite)
 
     is_active = flow.BoolParam(True)
@@ -1010,9 +1089,6 @@ class SiteMap(flow.Map):
         row["Site"] = item.name()
         if not item.configured.get():
             row["Site"] += " ⚠️"
-    
-    def _fill_row_style(self, style, item, row):
-        style["activate_oid"] = item.configuration.oid()
 
 
 class ConfigureWorkingSite(ConfigureSite):
@@ -1071,6 +1147,9 @@ class GotoCurrentSiteQueue(flow.Action):
     def needs_dialog(self):
         return False
     
+    def allow_context(self, context):
+        return context and context.endswith('.details')
+    
     def run(self, button):
         current_site = self.root().project().get_current_site()
         return self.get_result(goto=current_site.queue.oid())
@@ -1223,8 +1302,11 @@ class SiteEnvironment(flow.Map):
         row['Variable'] = item.variable.get()
         row['Value'] = item.value.get()
 
+from .users import PresetValue
 
-class SiteJobsPoolNames(flow.values.ChoiceValue):
+class SiteJobsPoolNames(PresetValue):
+
+    DEFAULT_EDITOR = 'choice'
     
     def choices(self):
         site = self.root().project().get_current_site()
@@ -1288,13 +1370,7 @@ class WorkingSite(Site):
             child_value.set(root_dir)
 
         elif child_value is self.configured:
-            self.configured.set(
-                (
-                    self.root_windows_folder.get()
-                    or self.root_linux_folder.get()
-                    or self.root_darwin_folder.get()
-                )
-            )
+            self.configured.set(bool(self.root_folder.get()))
         elif child_value is self.ordered_site_names:
             self.ordered_site_names.set(self._compute_ordered_site_names())
     
@@ -1321,6 +1397,9 @@ class WorkingSite(Site):
     
     def get_jobs(self, type=None, status=None):
         return self.get_queue().jobs(type=type, status=status)
+    
+    def count_jobs(self, type=None, status=None):
+        return self.get_queue().count(type=type, status=status)
 
 
 class ClearSiteQueues(flow.Action):
@@ -1394,6 +1473,32 @@ class ConfigureExchangeSite(ConfigureSite):
         site.bucket_name.set(self.bucket_name.get())
 
 
+class TestConnectionAction(flow.Action):
+
+    _exchange = flow.Parent()
+
+    def needs_dialog(self):
+        return True
+
+    def get_buttons(self):
+        self.message.set('Clik the button to test the connection.')
+        return ['Test']
+
+    def run(self, button):
+        ret = self._exchange.sync_manager.check_connection()
+
+        if ret is not None:
+            self.message.set(
+                '<font color=red>Connection error:</font><br>'
+                f'<pre>{ret}</pre>'
+            )
+        else:
+            self.message.set(
+                'Connection looks <b>OK</b>'
+            )
+        return self.get_result(close=False)
+
+
 class ExchangeSite(Site):
 
     ICON = ('icons.libreflow', 'exchange')
@@ -1402,8 +1507,12 @@ class ExchangeSite(Site):
     server_login = flow.Param("")
     server_password = flow.Param("").ui(editor='password')
     bucket_name = flow.Param('')
+    enable_tls = flow.BoolParam(False)
+
+    sync_manager = flow.Child(SyncManager).ui(hidden=True)
 
     configuration = flow.Child(ConfigureExchangeSite)
+    test_connection = flow.Child(TestConnectionAction)
 
     def compute_child_value(self, child_value):
         if child_value is self.configured:
@@ -1663,7 +1772,7 @@ class RequestRevisions(flow.Action):
 
 class SiteMultiSelection(flow.Object):
     
-    source_site = flow.Param(None, ActiveSiteAutoSelectChoiceValue).watched()
+    source_site = flow.Param('Auto select', ActiveSiteAutoSelectChoiceValue).watched()
     target_sites = flow.Param([], ActiveSitesMultichoiceValue)
     
     def child_value_changed(self, child_value):
@@ -1725,7 +1834,7 @@ class MultiRequestRevisions(RequestRevisions):
     revisions = flow.Child(MultiRequestedRevisions).ui(expanded=True)
     
     def get_buttons(self):
-        self.sites.source_site.set(self.root().project().get_current_site().name())
+        self.sites.source_site.revert_to_default()
         target_sites_choices = self.sites.target_sites.choices()
         
         if not target_sites_choices:
@@ -1800,7 +1909,11 @@ class UploadRevision(flow.Action):
         return self._revision.get_sync_status(exchange=True) == 'Available'
     
     def allow_context(self, context):
-        return context and not self._revision.is_working_copy()
+        return (
+            context
+            and not self._revision.is_working_copy()
+            and self._revision.get_sync_status() == 'Available'
+        )
     
     def get_buttons(self):  
         self.message.set((
@@ -1819,11 +1932,11 @@ class UploadRevision(flow.Action):
             job_type='Upload',
             init_status='WAITING',
             emitter_oid=self._revision.oid(),
-            user=self.root().project().get_user(),
+            user=self.root().project().get_user_name(),
             studio=current_site.name(),
         )
-        process_jobs_action = self.root().project().admin.process_jobs
-        process_jobs_action._process(job)
+        sync_manager = self.root().project().get_sync_manager()
+        sync_manager.process(job)
         
         self._revisions.touch()
 
@@ -1849,11 +1962,11 @@ class DownloadRevision(flow.Action):
             job_type='Download',
             init_status='WAITING',
             emitter_oid=self._revision.oid(),
-            user=self.root().project().get_user(),
+            user=self.root().project().get_user_name(),
             studio=current_site.name(),
         )
         self._revision.set_sync_status('Requested', current_site.name())
-        process_jobs_action = self.root().project().admin.process_jobs
-        process_jobs_action._process(job)
+        sync_manager = self.root().project().get_sync_manager()
+        sync_manager.process(job)
         
         self._revisions.touch()
